@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2021 Genome Research Ltd.
+ * Copyright (c) 2017-2023 Genome Research Ltd.
  * Author(s): James Bonfield
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
 
 #include "config.h"
 
-#ifdef HAVE_AVX2
+#if defined(__x86_64__) && defined(HAVE_AVX2)
 
 #include <stdint.h>
 #include <stdlib.h>
@@ -157,7 +157,8 @@ unsigned char *rans_compress_O0_32x16_avx2(unsigned char *in,
         goto empty;
 
     // Compute statistics
-    hist8(in, in_size, F);
+    if (hist8(in, in_size, F) < 0)
+        return NULL;
 
     // Normalise so frequences sum to power of 2
     uint32_t fsum = in_size;
@@ -525,7 +526,6 @@ unsigned char *rans_uncompress_O0_32x16_avx2(unsigned char *in,
         Rv2 = _mm256_add_epi32(
                   _mm256_mullo_epi32(
                       _mm256_srli_epi32(Rv2,TF_SHIFT), fv2), bv2);
-
 #ifdef __clang__
         // Protect against running off the end of in buffer.
         // We copy it to a worst-case local buffer when near the end.
@@ -648,6 +648,7 @@ unsigned char *rans_uncompress_O0_32x16_avx2(unsigned char *in,
         Yv3 = _mm256_or_si256(Yv3, Vv3);
         Vv4 = _mm256_permutevar8x32_epi32(Vv4, idx4);
         Yv4 = _mm256_or_si256(Yv4, Vv4);
+        sp += _mm_popcnt_u32(imask4);
 
 #ifndef __clang__
         // 26% faster here than above for gcc10, but former location is
@@ -662,8 +663,6 @@ unsigned char *rans_uncompress_O0_32x16_avx2(unsigned char *in,
         }
 #endif
 
-        sp += _mm_popcnt_u32(imask4);
-
         // R[z] = c ? Y[z] : R[z];
         Rv3 = _mm256_blendv_epi8(Rv3, Yv3, renorm_mask3);
         Rv4 = _mm256_blendv_epi8(Rv4, Yv4, renorm_mask4);
@@ -1078,7 +1077,7 @@ unsigned char *rans_uncompress_O1_32x16_avx2(unsigned char *in,
         uint32_t u_freq_sz, c_freq_sz;
         cp += var_get_u32(cp, cp_end, &u_freq_sz);
         cp += var_get_u32(cp, cp_end, &c_freq_sz);
-        if (c_freq_sz >= cp_end - cp - 16)
+        if (c_freq_sz > cp_end - cp)
             goto err;
         tab_end = cp + c_freq_sz;
         if (!(c_freq = rans_uncompress_O0_4x16(cp, c_freq_sz, NULL,
@@ -1123,7 +1122,7 @@ unsigned char *rans_uncompress_O1_32x16_avx2(unsigned char *in,
     union {
         unsigned char tbuf[32][32];
         uint64_t tbuf64[32][4];
-    } u;
+    } u  __attribute__((aligned(32)));
     unsigned int tidx = 0;
 
     if (0) {
@@ -1643,4 +1642,7 @@ unsigned char *rans_uncompress_O1_32x16_avx2(unsigned char *in,
 
     return NULL;
 }
+#else  // HAVE_AVX2
+// Prevent "empty translation unit" errors when building without AVX2
+const char *rANS_static32x16pr_avx2_disabled = "No AVX2";
 #endif // HAVE_AVX2
